// $Date:: 2014-07-27 02:52:58 +0900#$
// $Rev: 11 $

#include "convert8to16_noopt.h"
#include <stdio.h>
#include <assert.h>

//#define PRINTTIME
#ifdef PRINTTIME
#include <stdio.h>
#include <mmsystem.h>
#pragma comment ( lib , "winmm.lib" )
static __int32 tcount = 0;
#endif

using namespace convert8to16_noopt;

static inline void SetPixcel(
	const uint8_t * const pSrc,
	const uint32_t nSrcPitch,
	uint16_t* const pDst,
	const uint32_t nDstPitch,
	const uint32_t nHeight,
	uint32_t * const pBuffer,
	uint32_t * const pHistgram);
static uint32_t xor128();

/**************************************************************************************************/
/* Convert8to16                                                                                      */
/**************************************************************************************************/
//==============================================================================
// constructor
//==============================================================================
Convert8to16::Convert8to16(
	PClip _child,
	IScriptEnvironment* env,
	const uint32_t span,
	const uint32_t ythr,
	const uint32_t cthr,
	const uint32_t mode,
	const bool debug) :
	GenericVideoFilter(_child),
	Span(span),
	YThr(ythr),
	CThr(cthr),
	Mode(mode),
	Debug(debug)
{
	this->BuffSize = 0;
	if (!this->Debug)
	{
		this->vi.width = this->vi.width * 2;
	}
}

//==============================================================================
// destructor
//==============================================================================
Convert8to16::~Convert8to16()
{
	if (this->BuffSize)
	{
		free(this->v1);
		free(this->v2);
		free(this->v3);
		free(this->v4);
	}


#ifdef PRINTTIME
	char strtime[256];
	_snprintf(strtime, sizeof(strtime), "%f\r\n", ((float)tcount) / 1000);
	OutputDebugString(strtime);
#endif // PRINTTIME
}

//==============================================================================
// GetFrame
//==============================================================================
PVideoFrame Convert8to16::GetFrame(
	int n,
	IScriptEnvironment * env)
{
#ifdef PRINTTIME
	tcount -= timeGetTime();
#endif // PRINTTIME

	PVideoFrame src = this->child->GetFrame(n, env);
	PVideoFrame dst = env->NewVideoFrame(this->vi);
	env->MakeWritable(&dst);

	SetPixcel(src, dst, PLANAR_Y);
	if (this->Debug)
	{
		memset(dst->GetWritePtr(PLANAR_U), 128, dst->GetPitch(PLANAR_U)*src->GetHeight(PLANAR_U));
		memset(dst->GetWritePtr(PLANAR_V), 128, dst->GetPitch(PLANAR_V)*src->GetHeight(PLANAR_V));
	}
	else
	{
		SetPixcel(src, dst, PLANAR_U);
		SetPixcel(src, dst, PLANAR_V);
		//memset(dst->GetWritePtr(PLANAR_U), 128, dst->GetPitch(PLANAR_U)*src->GetHeight(PLANAR_U));
		//memset(dst->GetWritePtr(PLANAR_V), 128, dst->GetPitch(PLANAR_V)*src->GetHeight(PLANAR_V));
	}

#ifdef PRINTTIME
	tcount += timeGetTime();
#endif // PRINTTIME
	return dst;
}

void Convert8to16::SetPixcel(
	PVideoFrame & src,
	PVideoFrame & dst,
	uint32_t nPlanar)
{
	const uint8_t* const pSrc = src->GetReadPtr(nPlanar);
	const uint32_t nSrcPitch = src->GetPitch(nPlanar);
	const uint32_t nHeight = src->GetHeight(nPlanar);
	const uint32_t nDstPitch = dst->GetPitch(nPlanar) / (this->Debug ? 1 : 2);
	const uint32_t nSpan = this->Span / ((nPlanar == PLANAR_Y) ? 1 : 2);

	if (this->BuffSize < nDstPitch * nHeight)
	{
		this->BuffSize = nDstPitch * nHeight;
		this->v1 = (int8_t *)realloc(this->v1, this->BuffSize * sizeof(int8_t));
		this->v2 = (BYTE *)realloc(this->v2, this->BuffSize * sizeof(BYTE));
		this->v3 = (uint16_t *)realloc(this->v3, this->BuffSize * sizeof(uint16_t));
		this->v4 = (BYTE *)realloc(this->v4, this->BuffSize * sizeof(BYTE));
	}

	memset(this->v1, 0x00, this->BuffSize);
	memset(this->v2, 0xFF, this->BuffSize);
	memset(this->v3, 0, this->BuffSize * sizeof(uint16_t));
	memset(this->v4, 0xFF, this->BuffSize);

	for (uint32_t i = 1; i < nHeight - 1; i++)
	{
		for (uint32_t j = 1; j < nDstPitch - 1; j++)
		{
#define SRC(a,b) ((int32_t)pSrc[nSrcPitch*(b)+(a)])
			int o, p, m;
			o = p = m = 0;
			for (int ii = -1; ii <= 1; ii++)
			{
				for (int jj = -1; jj <= 1; jj++)
				{
					switch (SRC(j, i) - SRC(j + jj, i + ii))
					{
					case 0: break;
					case 1: m++; break;
					case -1: p++; break;
					default: o++; break;
					}
				}
			}
#undef SRC
#define DST(a,b) (this->v1[nDstPitch*(b)+(a)])
#define FLG(a,b) (this->v2[nDstPitch*(b)+(a)])
			if (o == 0 && (this->Mode != 1 || (m == 0 || p == 0)))
			{
				int v;
				switch (p - m)
				{
				case -4: v = -0x70; break;
				case -3: v = -0x60; break;
				case -2: v = -0x40; break;
				case -1: v = -0x20; break;
				case 0: v = 0x00; break;
				case 1: v = 0x20; break;
				case 2: v = 0x40; break;
				case 3: v = 0x60; break;
				case 4: v = 0x70; break;
				default: v = (p) ? 0x7F : -0x7F;
				}
				DST(j, i) = v;
				FLG(j, i) = 0;
			}
			else
			{
				DST(j, i) = 0x00;
				FLG(j, i) = 0xFF;
			}
#undef FLG
#undef DST
		}
	}
	const int divc0 = (1 << 24) / ((2 * nSpan + 1)*(2 * nSpan + 1)) + 1;
	const int divc1 = divc0 >> 8;
	const int divc2 = divc0 & 0xFF;
	const uint32_t thr = ((nPlanar == PLANAR_Y) ? this->YThr : this->CThr);
	const int area = ((2 * nSpan + 1)*(2 * nSpan + 1));
#define SRC(a,b) ((uint16_t)pSrc[nSrcPitch*(b)+(a)])
#define SUB(a,b) (this->v1[nDstPitch*(b)+(a)])
#define DST(a,b) (this->v3[nDstPitch*(b)+(a)])
#define FLG(a,b) (this->v2[nDstPitch*(b)+(a)])
#define FLG2(a,b) (this->v4[nDstPitch*(b)+(a)])
	for (uint32_t i = 0; i < nHeight; i++)
	{
		for (uint32_t j = 0; j < nDstPitch; j++)
		{
			DST(j, i) = (SRC(j, i) << 8) + SUB(j, i);
			FLG2(j, i) = FLG(j, i) | 0x5F;
		}
	}
	for (uint32_t i = nSpan; i < nHeight - nSpan; i++)
	{
		for (uint32_t j = nSpan; j < nDstPitch - nSpan; j++)
		{
			uint32_t o = 0;
			uint32_t a = 0;
			for (int32_t ii = -(int32_t)nSpan; ii <= (int32_t)nSpan; ii++)
			{
				for (int32_t jj = -(int32_t)nSpan; jj <= (int32_t)nSpan; jj++)
				{
					if (FLG(j + jj, i + ii)) o++;
					a += (SRC(j + jj, i + ii) << 8) + SUB(j + jj, i + ii);
				}
			}
			if (o <= thr)
			{
				DST(j, i) = (a * divc1 + ((a * divc2) >> 8)) >> 16;
				assert(abs((int)DST(j, i) - (int)a / area) <= 1); // 2ȏȂƂ`FbN
				//DST(j, i) = a / area;
				//DST(j, i) = (SRC(j, i) << 8) + SUB(j, i);
				FLG2(j, i) = 0;
			}
			else
			{
				DST(j, i) = (SRC(j, i) << 8) + SUB(j, i);
				FLG2(j, i) = FLG(j, i) | 0x5F;
			}
		}
	}
#undef SRC
#undef SUB
#undef DST
#undef FLG
#undef FLG2
	if (this->Debug)
	{
		memcpy(dst->GetWritePtr(nPlanar), this->v4, nDstPitch * nHeight);
	}
	else
	{
		memcpy(dst->GetWritePtr(nPlanar), this->v3, nDstPitch * nHeight*sizeof(uint16_t));
	}
	return;
}

static uint32_t xor128()
{
	static uint32_t x = 123456789, y = 362436069, z = 521288629, w = 88675123;
	uint32_t t = x ^ (x << 11);
	x = y;
	y = z;
	z = w;
	w = (w ^ (w >> 19)) ^ (t ^ (t >> 8));
	return w;
}
