// $Date:: 2014-07-27 02:52:58 +0900#$
// $Rev: 11 $

#include "convert8to16_nosse.h"
#include <stdio.h>
#include <malloc.h>
#include <list>

#define ALGN 8
//#define PRINTTIME
#ifdef PRINTTIME
#include <stdio.h>
#include <mmsystem.h>
#pragma comment ( lib , "winmm.lib" )
static __int32 tcount = 0;
#endif

using namespace convert8to16_nosse;

static inline void SetPixcel(
	const uint8_t * const pSrc,
	const uint32_t nSrcPitch,
	uint16_t* const pDst,
	const uint32_t nPitch,
	const uint32_t nHeight,
	uint32_t * const pBuffer,
	uint32_t * const pHistgram);
static inline void hadd(
	const uint32_t span,
	const uint32_t pitch,
	const uint16_t * const src,
	const uint8_t * const flg,
	uint32_t * const dest,
	uint16_t * const flgd);
static inline uint16_t edige(const int16_t d);
static inline uint16_t edigeinv(const uint16_t d);
static inline void hcalcedige(
	const uint16_t * const p,
	const uint16_t * const n,
	const uint32_t nWidth,
	uint16_t * const v0,
	uint16_t * const v1);

/**************************************************************************************************/
/* Convert8to16                                                                                      */
/**************************************************************************************************/
//==============================================================================
// constructor
//==============================================================================
Convert8to16::Convert8to16(
	PClip _child,
	IScriptEnvironment* env,
	const uint32_t span,
	const uint32_t ythr,
	const uint32_t cthr,
	const uint32_t mode,
	const bool debug) :
GenericVideoFilter(_child),
	Span(span),
	YThr(ythr),
	CThr(cthr),
	Mode(mode),
	Debug(debug)
{
	this->BuffSize = 0;
	if (!this->Debug)
	{
		this->vi.width = this->vi.width * 2;
	}
}

//==============================================================================
// destructor
//==============================================================================
Convert8to16::~Convert8to16()
{
	if (this->BuffSize)
	{
		free(this->v1);
		free(this->v2);
		free(this->v3);
		free(this->v4);
		free(this->v5);
		free(this->v6);
		free(this->v7);
	}

#ifdef PRINTTIME
	char strtime[256];
	_snprintf(strtime, sizeof(strtime), "%f\r\n", ((float)tcount) / 1000);
	OutputDebugString(strtime);
#endif // PRINTTIME
}

//==============================================================================
// GetFrame
//==============================================================================
PVideoFrame Convert8to16::GetFrame(
	int n,
	IScriptEnvironment * env)
{
#ifdef PRINTTIME
	tcount -= timeGetTime();
#endif // PRINTTIME

	PVideoFrame src = this->child->GetFrame(n, env);
	PVideoFrame dst = env->NewVideoFrame(this->vi);
	env->MakeWritable(&dst);

	SetPixcel(src, dst, PLANAR_Y);
	if (this->Debug)
	{
		memset(dst->GetWritePtr(PLANAR_U), 128, dst->GetPitch(PLANAR_U)*src->GetHeight(PLANAR_U));
		memset(dst->GetWritePtr(PLANAR_V), 128, dst->GetPitch(PLANAR_V)*src->GetHeight(PLANAR_V));
	}
	else
	{
		SetPixcel(src, dst, PLANAR_U);
		SetPixcel(src, dst, PLANAR_V);
		//memset(dst->GetWritePtr(PLANAR_U), 128, dst->GetPitch(PLANAR_U)*src->GetHeight(PLANAR_U));
		//memset(dst->GetWritePtr(PLANAR_V), 128, dst->GetPitch(PLANAR_V)*src->GetHeight(PLANAR_V));
	}
#ifdef PRINTTIME
	tcount += timeGetTime();
#endif // PRINTTIME
	return dst;
}

void Convert8to16::SetPixcel(
	PVideoFrame & src,
	PVideoFrame & dst,
	uint32_t nPlanar)
{
	const uint32_t nHeight = src->GetHeight(nPlanar);
	const uint32_t nSrcPitch = src->GetPitch(nPlanar);
	const uint32_t nPitch = nSrcPitch + ((~nSrcPitch + 1)&(ALGN - 1));
	const uint32_t nWidth = this->vi.width / (this->Debug ? 1 : 2) / ((nPlanar == PLANAR_Y) ? 1 : 2);
	const uint32_t nSpan = this->Span / ((nPlanar == PLANAR_Y) ? 1 : 2);

	if (this->BuffSize < nPitch * nHeight)
	{
		this->BuffSize = nPitch * nHeight;
		this->v1 = (uint16_t *)realloc(this->v1, this->BuffSize * sizeof(uint16_t));
		this->v2 = (BYTE *)realloc(this->v2, this->BuffSize * sizeof(BYTE));
		this->v3 = (uint16_t *)realloc(this->v3, nPitch * sizeof(uint16_t));
		this->v4 = (uint16_t *)realloc(this->v4, nPitch * sizeof(uint16_t));
		this->v5 = (uint32_t *)realloc(this->v5, nPitch * sizeof(uint32_t));
		this->v6 = (uint32_t *)realloc(this->v6, nPitch * (2 * nSpan + 2) * sizeof(uint32_t));
		this->v7 = (uint16_t *)realloc(this->v7, nPitch * (2 * nSpan + 2) * sizeof(uint16_t));
	}

	{
		const uint8_t* const pSrc = src->GetReadPtr(nPlanar);
		for (uint32_t i = 0; i < nHeight; i++)
		{
			const uint8_t * s = pSrc + nSrcPitch * i;
			uint16_t * d1 = this->v1 + nPitch * i;
			for (uint32_t j = 0; j < nWidth; j++) d1[j] = s[j];
		}
	}


	{
		//const int16_t v[9] = { 0, 0x20, 0x40, 0x60, 0x70, 0x80, 0x80, 0x80, 0x80 };
		const int16_t v[9] = { 0, 0x20, 0x40, 0x60, 0x70, 0x7F, 0x7F, 0x7F, 0x7F };
		uint16_t * pd = this->v1;
		uint8_t * pf = this->v2;

		memset(this->v3, 0, nPitch * sizeof(uint16_t));
		hcalcedige(pd, pd + nPitch, nWidth, this->v3, this->v4);

		for (uint32_t j = 0; j < nWidth; j++) pd[j] <<= 8;
		memset(pf, 0xFF, nPitch);

		pd += nPitch, pf += nPitch;
		for (uint32_t i = 1; i < nHeight - 1; i++, pd += nPitch, pf += nPitch)
		{
			hcalcedige(pd, pd + nPitch, nWidth, this->v3, this->v4);
			for (uint32_t j = 0; j < nWidth; j++) pd[j] <<= 8;

			pf[0] = 0xFF;
			if (this->Mode == 1)
			{
				for (uint32_t j = 1; j < nWidth - 1; j++)
				{
					const uint16_t o = this->v4[j];
					const uint8_t f = ((o & 0xF00) == 0 && ((o & 0x0F0) == 0 || (o & 0x00F) == 0)) ? 0 : 0xFF;
					pf[j] = f;
					if (!f)
					{
						int16_t vv = 0;
						vv += v[o & 0xF];
						vv -= v[(o >> 4) & 0xF];
						pd[j] += vv;
					}
				}
			}
			else
			{
				for (uint32_t j = 1; j < nWidth - 1; j++)
				{
					const uint16_t o = this->v4[j];
					const uint8_t f = ((o & 0xF00) == 0) ? 0 : 0xFF;
					pf[j] = f;
					if (!f)
					{
						int16_t vv = 0;
						vv += v[o & 0xF];
						vv -= v[(o >> 4) & 0xF];
						pd[j] += vv;
					}
				}
			}
			pf[nWidth - 1] = 0xFF;
		}

		for (uint32_t j = 0; j < nWidth; j++) pd[j] <<= 8;
		memset(pf, 0xFF, nPitch);
	}

	std::list<uint32_t*> vv6;
	std::list<uint16_t*> vv7;

	memset(this->v4, 0, nPitch * sizeof(uint16_t));
	memset(this->v5, 0, nPitch * sizeof(uint32_t));
	memset(this->v6, 0, nPitch * (2 * nSpan + 2) * sizeof(uint32_t));
	memset(this->v7, 0, nPitch * (2 * nSpan + 2) * sizeof(uint16_t));

	for (uint32_t i = 0; i < 2 * nSpan + 1; i++)
	{
		vv6.push_back(this->v6 + nPitch*i);
		vv7.push_back(this->v7 + nPitch*i);
		hadd(
			nSpan,
			nWidth,
			this->v1 + nPitch*i,
			this->v2 + nPitch*i,
			vv6.back(),
			vv7.back());
		const uint32_t * const vvn5 = vv6.back();
		const uint16_t * const vvn8 = vv7.back();
		for (uint32_t j = nSpan; j < nWidth - nSpan; j++)
		{
			this->v5[j] += vvn5[j];
			this->v4[j] += vvn8[j];
		}
	}
	vv6.push_back(this->v6 + nPitch*(2 * nSpan + 1));
	vv7.push_back(this->v7 + nPitch*(2 * nSpan + 1));

	const int divc0 = (1 << 24) / ((2 * nSpan + 1)*(2 * nSpan + 1)) + 1;
	const int divc1 = divc0 >> 8;
	const int divc2 = divc0 & 0xFF;
	const int thr = ((nPlanar == PLANAR_Y) ? this->YThr : this->CThr);

	for (uint32_t i = nSpan + 1; i < nHeight - nSpan; i++)
	{
		hadd(
			nSpan,
			nWidth,
			this->v1 + nPitch*(i + nSpan),
			this->v2 + nPitch*(i + nSpan),
			vv6.back(),
			vv7.back());

		if (!this->Debug)
		{
			const uint16_t * const vvp8 = vv7.front();
			const uint16_t * const vvn8 = vv7.back();
			const uint32_t * const vvp5 = vv6.front();
			const uint32_t * const vvn5 = vv6.back();
			uint16_t * const vn1 = this->v1 + nPitch*i;
			for (uint32_t j = nSpan; j < nWidth - nSpan; j++)
			{
				uint16_t o = this->v4[j];
				o -= vvp8[j];
				o += vvn8[j];
				this->v4[j] = o;
				uint32_t a = this->v5[j];
				a -= vvp5[j];
				a += vvn5[j];
				this->v5[j] = a;

				if (o <= thr)
				{
					// ZϘaŋߎ
					vn1[j] = (a * divc1 + ((a * divc2) >> 8)) >> 16;
					//vn1[j] = a / ((2 * nSpan + 1)*(2 * nSpan + 1));
				}
			}
		}
		else
		{
			const uint16_t * const vvp8 = vv7.front();
			const uint16_t * const vvn8 = vv7.back();
			uint8_t * const vn2 = this->v2 + nPitch*i;
			for (uint32_t j = nSpan; j < nWidth - nSpan; j++)
			{
				uint16_t o = this->v4[j];
				o -= vvp8[j];
				o += vvn8[j];
				this->v4[j] = o;

				if (o <= thr)
				{
					vn2[j] ^= vn2[j]; // = 0;
				}
				else
				{
					vn2[j] |= 0x5F;
				}
			}
		}

		vv6.push_back(vv6.front());
		vv6.pop_front();
		vv7.push_back(vv7.front());
		vv7.pop_front();
	}

	{
		BYTE * const dstp = dst->GetWritePtr(nPlanar);
		BYTE * const srcp = this->Debug ? (BYTE*)this->v2 : (BYTE*)this->v1;
		const uint32_t dpitch = dst->GetPitch(nPlanar);
		const uint32_t spitch = nPitch * (this->Debug ? 1 : 2);
		for (uint32_t i = 0; i < nHeight; i++)
		{
			memcpy(dstp + dpitch * i, srcp + spitch * i, this->vi.width);
		}
	}
	return;
}

static inline void hadd(
	const uint32_t span,
	const uint32_t pitch,
	const uint16_t * const src,
	const uint8_t * const flg,
	uint32_t * const dest,
	uint16_t * const flgd)
{
	uint32_t a = 0;
	uint16_t o = 0;
	for (uint32_t j = 0; j < 2 * span + 1; j++)
	{
		if (flg[j]) o++;
		a += src[j];
	}
	dest[span] = a;
	flgd[span] = o;
	for (uint32_t j = span + 1; j < pitch - span; j++)
	{
		if (flg[j - span - 1]) o--;
		if (flg[j + span]) o++;
		a -= src[j - span - 1];
		a += src[j + span];
		dest[j] = a;
		flgd[j] = o;
	}
	for (uint32_t j = 0; j < span + 1; j++)
		flgd[j] = 0x00FF;
	for (uint32_t j = pitch - span; j < pitch; j++)
		flgd[j] = 0x00FF;
}

static inline uint16_t edige(const int16_t d)
{
	switch (d)
	{
	case 0: return 0x000;
	case -1: return 0x001;
	case 1: return 0x010;
	default: return 0x100;
	}
	//uint16_t o; o^=o;
	//o |= (d == 0) ? 0x1000 : 0;
	//o |= (d == -1) ? 0x0001 : 0;
	//o |= (d == 1) ? 0x0010 : 0;
	//o |= (o == 0) ? 0x0100 : 0;
	//return o;
}

static inline void hcalcedige(
	const uint16_t * const p,
	const uint16_t * const n,
	const uint32_t nWidth,
	uint16_t * const v0,
	uint16_t * const v1)
{
	const int16_t * const pp = (int16_t*)p;
	const int16_t * const nn = (int16_t*)n;
	uint16_t t0 = 0, t1 = 0, t2 = 0, t3 = 0, ap = 0, an = 0;
	for (uint32_t i = 0; i < nWidth - 1; i++)
	{
		ap ^= ap;
		an = v0[i];
		ap += t0;
		an += edigeinv(t3);
		an += t2;
		t0 = edige(pp[i] - nn[i + 1]);
		t1 = edige(pp[i] - nn[i]);
		t2 = edige(pp[i + 1] - nn[i]);
		t3 = edige(pp[i] - pp[i + 1]);
		ap += t1;
		ap += t2;
		an += t0;
		an += t1;
		an += t3;
		v0[i] = edigeinv(ap);
		v1[i] = an;
	}
	v1[0] = v1[nWidth - 1] = 0x0800;
}

static inline uint16_t edigeinv(const uint16_t d)
{
	return ((d & 0xFF00) | ((d >> 4) & 0x000F) | ((d << 4) & 0x00F0));
}
