// $Date:: 2014-07-27 02:52:58 +0900#$
// $Rev: 11 $

#include "convert8to16_sse4.h"
#include <stdio.h>
#include <list>
#include <malloc.h>    // aligned_malloc
#include <tmmintrin.h> // SSE3
#include <smmintrin.h> // SSE4.1

#define ALGN 16
//#define PRINTTIME
#ifdef PRINTTIME
#include <stdio.h>
#include <mmsystem.h>
#pragma comment ( lib , "winmm.lib" )
static __int32 tcount = 0;
#endif

using namespace convert8to16_sse4;

static inline void SetPixcel(
	const uint8_t * const pSrc,
	const uint32_t nSrcPitch,
	uint16_t* const pDst,
	const uint32_t nPitch,
	const uint32_t nHeight,
	uint32_t * const pBuffer,
	uint32_t * const pHistgram);
static inline void hadd(
	const uint32_t span,
	const uint32_t pitch,
	const uint16_t * const src,
	const uint16_t * const flg,
	uint32_t * const dest,
	uint16_t * const flgd);
static inline __m128i edige(const __m128i d);
static inline __m128i edigeinv(const __m128i d);
static inline void hcalcedige(
	const __m128i * const p,
	const __m128i * const n,
	const uint32_t nWidth,
	__m128i * const v0,
	__m128i * const v1);

/**************************************************************************************************/
/* Convert8to16                                                                                      */
/**************************************************************************************************/
//==============================================================================
// constructor
//==============================================================================
Convert8to16::Convert8to16(
	PClip _child,
	IScriptEnvironment* env,
	const uint32_t span,
	const uint32_t ythr,
	const uint32_t cthr,
	const uint32_t mode,
	const bool debug) :
GenericVideoFilter(_child),
	Span(span),
	YThr(ythr),
	CThr(cthr),
	Mode(mode),
	Debug(debug)
{
	if (this->vi.width % ALGN) env->ThrowError("equalize: Width should be a multiple of 16.");

	this->BuffSize = 0;
	if (!this->Debug)
	{
		this->vi.width = this->vi.width * 2;
	}
}

//==============================================================================
// destructor
//==============================================================================
Convert8to16::~Convert8to16()
{
	if (this->BuffSize)
	{
		_aligned_free(this->v1);
		_aligned_free(this->v2);
		_aligned_free(this->v3);
		_aligned_free(this->v4);
		_aligned_free(this->v5);
		_aligned_free(this->v6);
		_aligned_free(this->v7);
	}

#ifdef PRINTTIME
	char strtime[256];
	_snprintf(strtime, sizeof(strtime), "%f\r\n", ((float)tcount) / 1000);
	OutputDebugString(strtime);
#endif // PRINTTIME
}

//==============================================================================
// GetFrame
//==============================================================================
PVideoFrame Convert8to16::GetFrame(
	int n,
	IScriptEnvironment * env)
{
#ifdef PRINTTIME
	tcount -= timeGetTime();
#endif // PRINTTIME

	PVideoFrame src = this->child->GetFrame(n, env);
	PVideoFrame dst = env->NewVideoFrame(this->vi);
	env->MakeWritable(&dst);

	SetPixcel(src, dst, PLANAR_Y);
	if (this->Debug)
	{
		memset(dst->GetWritePtr(PLANAR_U), 128, dst->GetPitch(PLANAR_U)*src->GetHeight(PLANAR_U));
		memset(dst->GetWritePtr(PLANAR_V), 128, dst->GetPitch(PLANAR_V)*src->GetHeight(PLANAR_V));
	}
	else
	{
		SetPixcel(src, dst, PLANAR_U);
		SetPixcel(src, dst, PLANAR_V);
		//memset(dst->GetWritePtr(PLANAR_U), 128, dst->GetPitch(PLANAR_U)*src->GetHeight(PLANAR_U));
		//memset(dst->GetWritePtr(PLANAR_V), 128, dst->GetPitch(PLANAR_V)*src->GetHeight(PLANAR_V));
	}
#ifdef PRINTTIME
	tcount += timeGetTime();
#endif // PRINTTIME
	return dst;
}

static inline void ssememset(__m128i * dst, int32_t val, uint32_t size)
{
	const __m128i c = _mm_set1_epi8(val);
	for (uint32_t i = 0; i < size >> 4; i++, dst++) _mm_store_si128(dst, c);
}

void Convert8to16::SetPixcel(
	PVideoFrame & src,
	PVideoFrame & dst,
	uint32_t nPlanar)
{
	const uint32_t nHeight = src->GetHeight(nPlanar);
	const uint32_t nSrcPitch = src->GetPitch(nPlanar);
	const uint32_t nPitch = nSrcPitch + ((~nSrcPitch + 1)&(ALGN - 1)) + ALGN;
	const uint32_t nWidth = this->vi.width / (this->Debug ? 1 : 2) / ((nPlanar == PLANAR_Y) ? 1 : 2);
	const uint32_t nSpan = this->Span / ((nPlanar == PLANAR_Y) ? 1 : 2);

	if (this->BuffSize < nPitch * nHeight)
	{
		this->BuffSize = nPitch * nHeight;
		this->v1 = (uint16_t *)_aligned_realloc(this->v1, this->BuffSize * sizeof(uint16_t), ALGN);
		this->v2 = (uint16_t *)_aligned_realloc(this->v2, this->BuffSize * sizeof(uint16_t), ALGN);
		this->v3 = (__m128i *)_aligned_realloc(this->v3, nPitch * sizeof(uint16_t), ALGN);
		this->v4 = (uint16_t *)_aligned_realloc(this->v4, nPitch * sizeof(uint16_t), ALGN);
		this->v5 = (uint32_t *)_aligned_realloc(this->v5, nPitch * sizeof(uint32_t), ALGN);
		this->v6 = (uint32_t *)_aligned_realloc(this->v6, nPitch * (2 * nSpan + 2) * sizeof(uint32_t), ALGN);
		this->v7 = (uint16_t *)_aligned_realloc(this->v7, nPitch * (2 * nSpan + 2) * sizeof(uint16_t), ALGN);
	}

	{
		for (uint32_t i = 0; i < nHeight; i++)
		{
			const __m128i* s = (const __m128i*)(src->GetReadPtr(nPlanar) + i * nSrcPitch);
			__m128i * d = (__m128i*)(this->v1 + i * nPitch);
			_mm_prefetch((const char*)s, _MM_HINT_NTA);
			for (uint32_t j = 0; j < nWidth / ALGN; j++)
			{
				_mm_prefetch((const char*)(s + 1), _MM_HINT_NTA);
				__m128i a = _mm_lddqu_si128(s++);
				_mm_store_si128(d++, _mm_cvtepu8_epi16(a));
				a = _mm_srli_si128(a, 8);
				_mm_store_si128(d++, _mm_cvtepu8_epi16(a));
			}
		}
	}

	{
		static const __m128i vvv = _mm_setr_epi8(
			0x00, 0x20, 0x40, 0x60, 0x70, 0x7F, 0x7F, 0x7F,
			0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F, 0x7F);
		static const __m128i m0 = _mm_set1_epi16(0xF00);
		static const __m128i m1 = _mm_set1_epi16(0x0F0);
		static const __m128i m2 = _mm_set1_epi16(0x00F);
		static const __m128i cz = _mm_setzero_si128();
		__m128i * pd = (__m128i*)this->v1;
		__m128i * pf = (__m128i*)this->v2;
		__m128i * pv4 = (__m128i*)this->v4;

		ssememset(this->v3, 0, nPitch * sizeof(uint16_t));
		hcalcedige(pd, pd + nPitch / ALGN * 2, nWidth, this->v3, pv4);

		for (uint32_t j = 0; j < nWidth / ALGN * 2; j++) _mm_store_si128(pd + j, _mm_slli_epi16(_mm_load_si128(pd + j), 8));
		ssememset(pf, 0xFF, nPitch * 2);

		pd += nPitch / ALGN * 2, pf += nPitch / ALGN * 2;
		for (uint32_t i = 1; i < nHeight - 1; i++, pd += nPitch / ALGN * 2, pf += nPitch / ALGN * 2)
		{
			hcalcedige(pd, pd + nPitch / ALGN * 2, nWidth, this->v3, pv4);
			for (uint32_t j = 0; j < nWidth / ALGN * 2; j++) _mm_store_si128(pd + j, _mm_slli_epi16(_mm_load_si128(pd + j), 8));
			if (this->Mode == 1)
			{
				for (uint32_t j = 0; j < nWidth / ALGN * 2; j++)
				{
					const __m128i o = _mm_load_si128(pv4 + j);
					const __m128i f = _mm_and_si128(_mm_or_si128(
						_mm_cmpeq_epi16(_mm_and_si128(o, m1), cz),
						_mm_cmpeq_epi16(_mm_and_si128(o, m2), cz)),
						_mm_cmpeq_epi16(_mm_and_si128(o, m0), cz));
					_mm_store_si128(pf + j, _mm_cmpeq_epi16(f, cz));
					__m128i vv = _mm_sub_epi16(
						_mm_shuffle_epi8(vvv, _mm_and_si128(o, m2)),
						_mm_shuffle_epi8(vvv, _mm_and_si128(_mm_srli_epi16(o, 4), m2)));
					vv = _mm_and_si128(vv, f);
					_mm_store_si128(pd + j, _mm_add_epi16(_mm_load_si128(pd + j), vv));
				}
			}
			else
			{
				for (uint32_t j = 0; j < nWidth / ALGN * 2; j++)
				{
					const __m128i o = _mm_load_si128(pv4 + j);
					const __m128i f = _mm_cmpeq_epi16(_mm_and_si128(o, m0), cz);
					_mm_store_si128(pf + j, _mm_cmpeq_epi16(f, cz));
					__m128i vv = _mm_sub_epi16(
						_mm_shuffle_epi8(vvv, _mm_and_si128(o, m2)),
						_mm_shuffle_epi8(vvv, _mm_and_si128(_mm_srli_epi16(o, 4), m2)));
					vv = _mm_and_si128(vv, f);
					_mm_store_si128(pd + j, _mm_add_epi16(_mm_load_si128(pd + j), vv));
				}
			}
			uint16_t * const pfptr = (uint16_t*)pf;
			pfptr[0]=pfptr[nWidth - 1] = 0xFF;
		}

		for (uint32_t j = 0; j < nWidth / ALGN * 2; j++)_mm_store_si128(pd + j, _mm_slli_epi16(_mm_load_si128(pd + j), 8));
		ssememset(pf, 0xFF, nPitch * 2);
	}

	std::list<uint32_t*> vv6;
	std::list<uint16_t*> vv7;

	ssememset((__m128i*)this->v4, 0, nPitch * sizeof(uint16_t));
	ssememset((__m128i*)this->v5, 0, nPitch * sizeof(uint32_t));
	ssememset((__m128i*)this->v6, 0, nPitch * (2 * nSpan + 2) * sizeof(uint32_t));
	ssememset((__m128i*)this->v7, 0, nPitch * (2 * nSpan + 2) * sizeof(uint16_t));

	for (uint32_t i = 0; i < 2 * nSpan + 1; i++)
	{
		vv6.push_back(this->v6 + nPitch*i);
		vv7.push_back(this->v7 + nPitch*i);
		hadd(
			nSpan,
			nWidth,
			this->v1 + nPitch*i,
			this->v2 + nPitch*i,
			vv6.back(),
			vv7.back());
		const __m128i * const vvn8 = (const __m128i*)vv7.back();
		const __m128i * const vvn5 = (const __m128i*)vv6.back();
		for (uint32_t j = 0, k = 0; j < nWidth / ALGN * 2; j++)
		{
			_mm_store_si128((__m128i*)this->v4 + j, _mm_add_epi16(
				_mm_load_si128((__m128i*)this->v4 + j),
				_mm_load_si128(vvn8 + j)));

			_mm_store_si128((__m128i*)this->v5 + k, _mm_add_epi32(
				_mm_load_si128((__m128i*)this->v5 + k),
				_mm_load_si128(vvn5 + k)));
			k++;
			_mm_store_si128((__m128i*)this->v5 + k, _mm_add_epi32(
				_mm_load_si128((__m128i*)this->v5 + k),
				_mm_load_si128(vvn5 + k)));
			k++;
		}
	}
	vv6.push_back(this->v6 + nPitch*(2 * nSpan + 1));
	vv7.push_back(this->v7 + nPitch*(2 * nSpan + 1));

	const int divc0 = (1 << 24) / ((2 * nSpan + 1)*(2 * nSpan + 1)) + 1;
	const __m128i divc1 = _mm_set1_epi32(divc0 >> 8);
	const __m128i divc2 = _mm_set1_epi32(divc0 & 0xFF);
	const int _thr = this->YThr + 1;//((nPlanar == PLANAR_Y) ? this->YThr : this->CThr) + 1;
	const __m128i thr = _mm_set1_epi16(_thr);
	BYTE * const dstp = dst->GetWritePtr(nPlanar);
	const uint32_t dpitch = dst->GetPitch(nPlanar);

	for (uint32_t i = nSpan + 1; i < nHeight - nSpan; i++)
	{
		hadd(
			nSpan,
			nWidth,
			this->v1 + nPitch*(i + nSpan),
			this->v2 + nPitch*(i + nSpan),
			vv6.back(),
			vv7.back());

		if (!this->Debug)
		{
			const __m128i * const vvp8 = (const __m128i*)vv7.front();
			const __m128i * const vvn8 = (const __m128i*)vv7.back();
			const __m128i * const vvp5 = (const __m128i*)vv6.front();
			const __m128i * const vvn5 = (const __m128i*)vv6.back();
			const __m128i * const vn1 = (__m128i*)(this->v1 + nPitch*i);
			__m128i * const dst = (__m128i*)(dstp + dpitch*i);
			for (uint32_t j = 0, k = 0; j < nWidth / ALGN * 2; j++, k += 2)
			{
				_mm_prefetch((const char*)(vn1 + j), _MM_HINT_NTA);

				__m128i o = _mm_load_si128((__m128i*)this->v4 + j);
				o = _mm_sub_epi16(o, _mm_load_si128(vvp8 + j));
				o = _mm_add_epi16(o, _mm_load_si128(vvn8 + j));
				_mm_store_si128((__m128i*)this->v4 + j, o);

				__m128i a0 = _mm_load_si128((__m128i*)this->v5 + k);
				__m128i a1 = _mm_load_si128((__m128i*)this->v5 + k + 1);
				a0 = _mm_sub_epi32(a0, _mm_load_si128(vvp5 + k));
				a1 = _mm_sub_epi32(a1, _mm_load_si128(vvp5 + k + 1));
				a0 = _mm_add_epi32(a0, _mm_load_si128(vvn5 + k));
				a1 = _mm_add_epi32(a1, _mm_load_si128(vvn5 + k + 1));
				_mm_store_si128((__m128i*)this->v5 + k, a0);
				_mm_store_si128((__m128i*)this->v5 + k + 1, a1);

				a0 = _mm_srli_epi32(_mm_add_epi32(_mm_mullo_epi32(a0, divc1),
					_mm_srli_epi32(_mm_mullo_epi32(a0, divc2), 8)), 16);
				a1 = _mm_srli_epi32(_mm_add_epi32(_mm_mullo_epi32(a1, divc1),
					_mm_srli_epi32(_mm_mullo_epi32(a1, divc2), 8)), 16);

				__m128i t = _mm_packus_epi32(a0, a1);
				_mm_storeu_si128(dst + j, _mm_blendv_epi8(_mm_load_si128(vn1 + j), t,
					_mm_cmpgt_epi16(thr, o)));
			}
		}
		else
		{
			static const __m128i c0 = _mm_set1_epi8(0x5F);
			static const __m128i cz = _mm_setzero_si128();
			const __m128i * const vvp8 = (const __m128i*)vv7.front();
			const __m128i * const vvn8 = (const __m128i*)vv7.back();
			__m128i * const vn2 = (__m128i*)(this->v2 + nPitch*i);
			__m128i * const dst = (__m128i*)(dstp + dpitch*i);
			for (uint32_t j = 0, k = 0; j < nWidth / ALGN; j++, k += 2)
			{
				_mm_prefetch((const char*)(vn2 + k), _MM_HINT_NTA);

				__m128i o0 = _mm_load_si128((__m128i*)this->v4 + k);
				__m128i o1 = _mm_load_si128((__m128i*)this->v4 + k + 1);
				o0 = _mm_sub_epi16(o0, _mm_load_si128(vvp8 + k));
				o1 = _mm_sub_epi16(o1, _mm_load_si128(vvp8 + k + 1));
				o0 = _mm_add_epi16(o0, _mm_load_si128(vvn8 + k));
				o1 = _mm_add_epi16(o1, _mm_load_si128(vvn8 + k + 1));
				_mm_store_si128((__m128i*)this->v4 + k, o0);
				_mm_store_si128((__m128i*)this->v4 + k + 1, o1);

				__m128i t = _mm_packs_epi16(_mm_cmpgt_epi16(thr, o0), _mm_cmpgt_epi16(thr, o1));
				_mm_storeu_si128(dst + j, _mm_blendv_epi8(_mm_or_si128(_mm_packs_epi16(
					_mm_load_si128(vn2 + k), _mm_load_si128(vn2 + k + 1)), c0), cz, t));
			}
		}

		vv6.push_back(vv6.front());
		vv6.pop_front();
		vv7.push_back(vv7.front());
		vv7.pop_front();
	}

	if (!this->Debug)
	{
		for (uint32_t i = 0; i < nSpan + 1; i++)
		{
			memcpy(dstp + dpitch * i, this->v1 + nPitch * i, nWidth*sizeof(uint16_t));
		}
		for (uint32_t i = nHeight - nSpan; i < nHeight; i++)
		{
			memcpy(dstp + dpitch * i, this->v1 + nPitch * i, nWidth*sizeof(uint16_t));
		}
	}
	else
	{
		for (uint32_t i = 0; i < nSpan + 1; i++)
		{
			BYTE * const d = dstp + dpitch * i;
			uint16_t * const s = this->v2 + nPitch * i;
			for (int j = 0; j < nWidth; j++)d[j] = s[j] | 0x5F;
		}
		for (uint32_t i = nHeight - nSpan; i < nHeight; i++)
		{
			BYTE * const d = dstp + dpitch * i;
			uint16_t * const s = this->v2 + nPitch * i;
			for (int j = 0; j < nWidth; j++)d[j] = s[j] | 0x5F;
		}
	}
	return;
}

static inline void hadd(
	const uint32_t span,
	const uint32_t pitch,
	const uint16_t * const src,
	const uint16_t * const flg,
	uint32_t * const dest,
	uint16_t * const flgd)
{
	static const __m128i cz = _mm_setzero_si128();
	static const __m128i m0 = _mm_set1_epi16(1);

	for (uint32_t i = span + 1; i < pitch - span; i += 8)
	{
		__m128i * df = (__m128i *)(flgd + i);
		__m128i * f0 = (__m128i *)(flg + i - span - 1);
		__m128i * f1 = (__m128i *)(flg + i + span);
		__m128i oo = _mm_sub_epi16(
			_mm_andnot_si128(_mm_cmpeq_epi16(_mm_lddqu_si128(f1), cz), m0),
			_mm_andnot_si128(_mm_cmpeq_epi16(_mm_lddqu_si128(f0), cz), m0));
		_mm_storeu_si128(df, oo);

		__m128i * dd = (__m128i *)(dest + i);
		__m128i * s0 = (__m128i *)(src + i - span - 1);
		__m128i * s1 = (__m128i *)(src + i + span);
		__m128i v0 = _mm_lddqu_si128(s0);
		__m128i v1 = _mm_lddqu_si128(s1);
		_mm_storeu_si128(dd, _mm_sub_epi32(
			_mm_cvtepu16_epi32(v1),
			_mm_cvtepu16_epi32(v0)));
		_mm_storeu_si128(dd + 1, _mm_sub_epi32(
			_mm_cvtepu16_epi32(_mm_srli_si128(v1,8)),
			_mm_cvtepu16_epi32(_mm_srli_si128(v0, 8))));
	}
	uint32_t a = 0;
	uint16_t o = 0;
	for (uint32_t i =0; i < 2 * span + 1; i++)
	{
		if (flg[i]) o++;
		a += src[i];
	}

	uint32_t j = 0;
	for (j = 0; j < span; j++)
	{
		dest[j] = src[j] * (2 * span + 1);
		flgd[j] = 0x00FF;
	}
	dest[j] = a;
	flgd[j] = o;
	j++;
	for (; j < pitch - span; j++)
	{
		a += (int32_t)dest[j];
		o += (int16_t)flgd[j];
		dest[j] = a;
		flgd[j] = o;
	}
	for (; j < pitch; j++)
	{
		dest[j] = src[j] * (2 * span + 1);
		flgd[j] = 0x00FF;
	}
}

static inline __m128i edige(const __m128i d)
{
	static const __m128i cz = _mm_set1_epi16(0x1000);
	static const __m128i cm = _mm_set1_epi16(0x0001);
	static const __m128i cp = _mm_set1_epi16(0x0010);
	static const __m128i co = _mm_set1_epi16(0x0100);
	static const __m128i vz = _mm_setzero_si128();
	static const __m128i vm = _mm_set1_epi16(-1);
	static const __m128i & vp = cm;

	__m128i o = vz;
	o = _mm_or_si128(o, _mm_and_si128(_mm_cmpeq_epi16(d, vz), cz)); //o |= (d == 0) ? 0x1000 : 0;
	o = _mm_or_si128(o, _mm_and_si128(_mm_cmpeq_epi16(d, vm), cm)); //o |= (d == -1) ? 0x0001 : 0;
	o = _mm_or_si128(o, _mm_and_si128(_mm_cmpeq_epi16(d, vp), cp)); //o |= (d == 1) ? 0x0010 : 0;
	o = _mm_or_si128(o, _mm_and_si128(_mm_cmpeq_epi16(o, vz), co)); //o |= (o == 0) ? 0x0100 : 0;
	return o;
}

static inline void hcalcedige(
	const __m128i * const p,
	const __m128i * const n,
	const uint32_t nWidth,
	__m128i * const v0,
	__m128i * const v1)
{
	const __m128i * const & pp0 = p;
	const __m128i * const & nn0 = n;
	const __m128i * const pp1 = (__m128i*)((uint16_t*)p + 1);
	const __m128i * const nn1 = (__m128i*)((uint16_t*)n + 1);
	int32_t pt0 = 0, pt2 = 0, pt3 = 0;
	for (uint32_t i = 0; i < nWidth / ALGN * 2; i++)
	{
		const __m128i vpp0 = _mm_load_si128(pp0 + i);
		const __m128i vnn0 = _mm_load_si128(nn0 + i);
		const __m128i vpp1 = _mm_lddqu_si128(pp1 + i);
		const __m128i vnn1 = _mm_lddqu_si128(nn1 + i);
		__m128i ap = _mm_setzero_si128();
		__m128i an = _mm_load_si128(v0 + i);
		__m128i t0 = edige(_mm_sub_epi16(vpp0, vnn1));
		__m128i t1 = edige(_mm_sub_epi16(vpp0, vnn0));
		__m128i t2 = edige(_mm_sub_epi16(vpp1, vnn0));
		__m128i t3 = edige(_mm_sub_epi16(vpp0, vpp1));

		ap = _mm_add_epi16(ap, _mm_insert_epi16(_mm_slli_si128(t0, 2), pt0, 0));
		an = _mm_add_epi16(an, edigeinv(_mm_insert_epi16(_mm_slli_si128(t3, 2), pt3, 0)));
		an = _mm_add_epi16(an, _mm_insert_epi16(_mm_slli_si128(t2,2),pt2,0));
		ap = _mm_add_epi16(ap, t1);
		ap = _mm_add_epi16(ap, t2);
		an = _mm_add_epi16(an, t0);
		an = _mm_add_epi16(an, t1);
		an = _mm_add_epi16(an, t3);
		_mm_store_si128(v0 + i, edigeinv(ap));
		_mm_store_si128(v1 + i, an);
		pt0 = _mm_extract_epi16(t0, 7);
		pt2 = _mm_extract_epi16(t2, 7);
		pt3 = _mm_extract_epi16(t3, 7);
	}
	uint16_t * const v1ptr = (uint16_t *)v1;
	v1ptr[0] = v1ptr[nWidth - 1] = 0x0800;
}

static inline __m128i edigeinv(const __m128i d)
{
	static const __m128i m0 = _mm_set1_epi16((short)0xFF00);
	static const __m128i m1 = _mm_set1_epi16(0x000F);
	return _mm_or_si128(_mm_or_si128(
		_mm_and_si128(d, m0),
		_mm_and_si128(_mm_srli_epi16(d, 4), m1)),
		_mm_slli_epi16(_mm_and_si128(d, m1), 4));
}
